;------------------------------------------------------------------------------+
; Wild text file viewer that allows viewing upto 204 lines of text.	       |
;									       |
; Usage: TEXTVIEW <textfile>						       |
;									       |
; Assumes that EOLs consist of 13, 10 (CR, LF) sequences.		       |
; Expands TAB (09h) characters into spaces (aligns at 8 characters).	       |
; Allows SMOOTH scrolling with Up, Down, PageUp and PageDown keys.	       |
; Home and End move to the begining and the end of file, respectively.	       |
; Esc key exits the program.						       |
;									       |
; TEXTVIEW has been written by Chris Dragan.				       |
; This and other free sources you can find at: http://chris.dragan.name/       |
;									       |
; To compile source of TEXTVIEW use Netwide Assembler:			       |
;    nasm -o textview.com textview.asm					       |
; Netwide Assembler's home page is: http://nasm.sourceforge.net/	       |
;------------------------------------------------------------------------------+

%define TEXT_COLOR	17h	; light gray (7) on blue (1)
%define LINE_WIDTH	80 * 2	; every of 80 fields consists of char and attr
%define NUM_LINES	204	; about 32768 (vid mem size) / 160 (line wid)

%define CRTC_PORT	0463h	; address in BIOS data area
%define CRTC_SCANLINE	08h
%define CRTC_ROWHI	0Ch
%define CRTC_ROWLO	0Dh

%define skip_instr	db 0BDh ; opcode of "mov bp, X" for skiping instr.
%define MakeWord(hi,lo) (((hi)<<8)+(lo))

org 100h

; Assume at startup: ax=bx=[sp]=0000h, cx=00FFh

;---------------- Load file ----------------------------------------------------

		; Set DTA
			xchg	ax, di		; Clear di
			mov	ax, 1A5Ch	; DOSfn 1Ah - Set DTA
			mov	dh, 0Fh 	; Pointer to text buffer
			mov	si, dx
			int	21h

		; Open first parameter's FCB
			cbw
			xchg	ax, dx
			int	21h		; DOSfn 0Fh - Open FCB

		; Set up FCB and load file
			mov	ah, 27h 	; DOSfn 27h - Read FCB
			int	21h		; read upto 7F80h bytes
			shl	cx, 7

;---------------- Draw file in video memory ------------------------------------

		; Prepare to "drawing" the file
			push	word 0B800h
			pop	es
			mov	ax, MakeWord(TEXT_COLOR, ' ')
			push	cx
			mov	cx, 4000h
		    rep stosw
			pop	cx
			xor	di, di
			mov	dl, NUM_LINES
_next_line:		add	bx, LINE_WIDTH

		; Load next character
_next_char:		lodsb
			cmp	al, 13		; CR
			jz	_draw_enter
			ja	_no_tab 	; TAB (09h) < CR (0Dh)

		; Draw character
			salc
			and	di, byte -16
			add	di, byte 16 -2
_no_tab:		stosw

		; Check if line exceeded
			cmp	di, bx
			jb	_draw_next_char

		; Skip the rest of the line
_find_enter:		dec	cx
			jz	_file_drawn
			lodsb
			cmp	al, 13		; CR
			jnz	_find_enter

		; Skip assumed LF after CR
_draw_enter:		inc	si
			dec	cx
			jz	_file_drawn

		; Move to next line
			mov	di, bx
			dec	dx
			loopnz	_next_line

		; Go to next character
			skip_instr
_draw_next_char:	loop	_next_char

		; Calculate last line address
_file_drawn:		mov	si, NUM_LINES-25
			sub	si, dx
			jnc	_many_lines
			xor	si, si
_many_lines:		shl	si, 4
			xor	bx, bx

		; Hide cursor
			mov	ah, 01h
			mov	ch, 20h
			int	10h

;---------------- Handle user input --------------------------------------------

		; Ignore keyboard buffer
_get_key:		mov	ax, 0C06h
			mov	dl, 0FFh
			int	21h

		; Set CRTC port address (assume CGA emulation)
			mov	dx, 03DAh

		; Get key code
			in	al, 60h

		; Check if Esc key
			cbw
			dec	ax	; al == Esc ?
			jnz	_dispatch_key

		; Reset video mode and exit
			mov	al, 3
			int	10h
			ret

		; Dispatch key (70=end/71=dn/72=pgdn/78=home/79=up/80=pgup)
_dispatch_key:		js	_get_key
			sub	al, 70
			jc	_get_key
			xchg	ax, cx
			shr	cx, 1
			jc	_arrow
			add	bp, 0B00h	;-- additional delay for page
			jnc	_display	;/
			shr	cx, 1
			jc	_page

		; Home and End
			mov	bx, si
			jnp	_display
			xor	bx, bx
			jmp	short _display

		; Wait for vertical retrace when arrow key pressed
_arrow: 		in	al, dx
			test	al, 08h
			jz	_arrow

		; PageUp and PageDown
_page:			jcxz	_move_up
			inc	bx
			cmp	bx, si
			jna	_display
_move_up:		or	bx, bx
			jz	_display
			dec	bx

;---------------- Move screen position -----------------------------------------

		; Wait for horizontal retrace
_display:		in	al, dx
			test	al, 01h
			jnz	_display

		; Update scanline position
			mov	dl, 0D4h
			mov	ax, MakeWord(0Fh, CRTC_SCANLINE)
			and	ah, bl
			out	dx, ax

		; Update row position
			push	bx
			and	bl, 0F0h
			lea	bx, [ebx+ebx*4]
			mov	ax, MakeWord(CRTC_ROWHI, CRTC_ROWLO)
			xchg	ah, bl
			out	dx, ax
			xchg	ax, bx
			out	dx, ax
			pop	bx

		; Loop the program
			jmp	short _get_key

;---------------- End of file --------------------------------------------------
